// The following code runs speed benchmarks for the different modules
// The first inference speed for each module is discarded due to lazy initialization of models

#include <chrono>
#include <iostream>
#include <string>
#include <vector>

#include "tf_sdk.h"

typedef std::chrono::high_resolution_clock Clock;

void benchmarkObjectDetection(const std::string& license);

void benchmarkFaceRecognitionLite(const std::string& license);

void benchmarkFaceRecognitionFull(const std::string& license);

void benchmarkFaceLandmarkDetection(const std::string& license);

void benchmarkLiveness(const std::string& license);

void benchMarkMaskDetection(const std::string& license);

int main() {
    // TODO: Replace with your license
    const std::string license = "${LICENSE_CODE}";

    std::cout << "Running speed benchmarks with 1280x720 image\n";

    benchmarkFaceRecognitionLite(license);
    benchmarkFaceRecognitionFull(license);
    benchmarkObjectDetection(license);
    benchmarkFaceLandmarkDetection(license);
    benchmarkLiveness(license);
    benchMarkMaskDetection(license);

    return 0;
}

void benchmarkFaceRecognitionLite(const std::string& license) {
    // Initialize the SDK
    Trueface::SDK tfSdk;
    bool valid = tfSdk.setLicense(license);

    if (!valid) {
        std::cout << "Error: the provided license is invalid." << std::endl;
        return;
    }

    // Load the image
    Trueface::ErrorCode errorCode = tfSdk.setImage("../images/headshot.jpg");
    if (errorCode != Trueface::ErrorCode::NO_ERROR) {
        std::cout << "Error: could not load the image" << std::endl;
        return;
    }

    // Discard first inference speed
    Trueface::Faceprint faceprint;
    bool foundFace;
    tfSdk.getLargestFaceFeatureVector(faceprint, foundFace);

    // Obtain the aligned chip
    Trueface::FaceBoxAndLandmarks faceBoxAndLandmarks;
    bool found;
    tfSdk.detectLargestFace(faceBoxAndLandmarks, found);

    uint8_t faceBuffer[112*112*3];
    tfSdk.extractAlignedFace(faceBoxAndLandmarks, faceBuffer);

    const size_t NUM_ITERATIONS = 100;
    double totalTime = 0;

    // Time the creation of the feature vector
    auto t1 = Clock::now();
    for (size_t i = 0; i < NUM_ITERATIONS; ++i) {
        tfSdk.getFaceFeatureVector(faceBuffer, faceprint);
    }
    auto t2 = Clock::now();
    totalTime = std::chrono::duration_cast<std::chrono::milliseconds>(t2 - t1).count();

    std::cout << "Average time face recognition lite: " << totalTime / NUM_ITERATIONS
              << " ms | " << NUM_ITERATIONS << " iterations" << std::endl;

}

void benchmarkFaceRecognitionFull(const std::string& license) {
    // Initialize the SDK
    Trueface::ConfigurationOptions options;
    options.frModel = Trueface::FacialRecognitionModel::FULL;

    Trueface::SDK tfSdk(options);
    bool valid = tfSdk.setLicense(license);

    if (!valid) {
        std::cout << "Error: the provided license is invalid." << std::endl;
        return;
    }

    // Load the image
    Trueface::ErrorCode errorCode = tfSdk.setImage("../images/headshot.jpg");
    if (errorCode != Trueface::ErrorCode::NO_ERROR) {
        std::cout << "Error: could not load the image" << std::endl;
        return;
    }

    // Discard first inference speed
    Trueface::Faceprint faceprint;
    bool foundFace;
    errorCode = tfSdk.getLargestFaceFeatureVector(faceprint, foundFace);

    // Obtain the aligned chip
    Trueface::FaceBoxAndLandmarks faceBoxAndLandmarks;
    bool found;
    tfSdk.detectLargestFace(faceBoxAndLandmarks, found);

    uint8_t faceBuffer[112*112*3];
    tfSdk.extractAlignedFace(faceBoxAndLandmarks, faceBuffer);

    if (errorCode != Trueface::ErrorCode::NO_ERROR) {
        std::cout << "Error: Could not benchmark full model, SDK not built with full model support" << std::endl;
        return;
    }

    const size_t NUM_ITERATIONS = 30;
    double totalTime = 0;

    // Time the creation of the feature vector
    auto t1 = Clock::now();
    for (size_t i = 0; i < NUM_ITERATIONS; ++i) {
        tfSdk.getFaceFeatureVector(faceBuffer, faceprint);
    }
    auto t2 = Clock::now();
    totalTime = std::chrono::duration_cast<std::chrono::milliseconds>(t2 - t1).count();

    std::cout << "Average time face recognition full: " << totalTime / NUM_ITERATIONS
    << " ms | " << NUM_ITERATIONS << " iterations" << std::endl;
}

void benchmarkObjectDetection(const std::string& license) {
    // Initialize the SDK with the fast object detection model
    Trueface::ConfigurationOptions options;
    options.objModel = Trueface::ObjectDetectionModel::FAST;

    Trueface::SDK tfSdk(options);
    bool valid = tfSdk.setLicense(license);

    if (!valid) {
        std::cout << "Error: the provided license is invalid." << std::endl;
        return;
    }

    // Load the image
    Trueface::ErrorCode errorCode = tfSdk.setImage("../images/bike.jpg");
    if (errorCode != Trueface::ErrorCode::NO_ERROR) {
        std::cout << "Error: could not load the image" << std::endl;
        return;
    }

    // Discard first inference speed
    std::vector<Trueface::BoundingBox> boundingBoxes;
    tfSdk.detectObjects(boundingBoxes);

    if (boundingBoxes.empty()) {
        std::cout << "Unable to detect objects\n";
        return;
    }

    const size_t NUM_ITERATIONS = 100;
    double totalTime = 0;

    // Time the creation of the feature vector
    auto t1 = Clock::now();
    for (size_t i = 0; i < NUM_ITERATIONS; ++i) {
        tfSdk.detectObjects(boundingBoxes);

    }
    auto t2 = Clock::now();
    totalTime = std::chrono::duration_cast<std::chrono::milliseconds>(t2 - t1).count();

    std::cout << "Average time object detection (fast mode): " << totalTime / NUM_ITERATIONS
              << " ms | " << NUM_ITERATIONS << " iterations" << std::endl;

}

void benchmarkFaceLandmarkDetection(const std::string& license) {
    // Initialize the SDK
    Trueface::ConfigurationOptions options;
    options.smallestFaceHeight = 40;
    Trueface::SDK tfSdk(options);
    bool valid = tfSdk.setLicense(license);

    if (!valid) {
        std::cout << "Error: the provided license is invalid." << std::endl;
        return;
    }

    // Load the image
    Trueface::ErrorCode errorCode = tfSdk.setImage("../images/headshot.jpg");
    if (errorCode != Trueface::ErrorCode::NO_ERROR) {
        std::cout << "Error: could not load the image" << std::endl;
        return;
    }

    // Discard first inference speed
    Trueface::FaceBoxAndLandmarks faceBoxAndLandmarks;
    bool found = false;
    tfSdk.detectLargestFace(faceBoxAndLandmarks, found);
    if (!found) {
        std::cout << "Unable to find face in image\n";
        return;
    }

    const size_t NUM_ITERATIONS = 100;
    double totalTime = 0;

    // Time the creation of the feature vector
    auto t1 = Clock::now();
    for (size_t i = 0; i < NUM_ITERATIONS; ++i) {
        tfSdk.detectLargestFace(faceBoxAndLandmarks, found);
    }
    auto t2 = Clock::now();
    totalTime = std::chrono::duration_cast<std::chrono::milliseconds>(t2 - t1).count();

    std::cout << "Average time face and landmark detection: " << totalTime / NUM_ITERATIONS
              << " ms | " << NUM_ITERATIONS << " iterations" << std::endl;

}


void benchmarkLiveness(const std::string& license) {
   // Initialize the SDK
    Trueface::ConfigurationOptions options;
    options.smallestFaceHeight = 40;
    Trueface::SDK tfSdk(options);
    bool valid = tfSdk.setLicense(license);

   if (!valid) {
       std::cout << "Error: the provided license is invalid." << std::endl;
       return;
   }

   // Load the image
   Trueface::ErrorCode errorCode = tfSdk.setImage("../images/headshot.jpg");
   if (errorCode != Trueface::ErrorCode::NO_ERROR) {
       std::cout << "Error: could not load the image" << std::endl;
       return;
   }

   // Discard first inference speed
   float leftEyeWinkScore, rightEyeWinkScore;
   errorCode = tfSdk.detectBlink(leftEyeWinkScore, rightEyeWinkScore);

   if (errorCode != Trueface::ErrorCode::NO_ERROR) {
       std::cout << "Unable to run blink model\n";
       return;
   }

   const size_t NUM_ITERATIONS = 100;
   double totalTime = 0;

   // Time the creation of the feature vector
   auto t1 = Clock::now();
   for (size_t i = 0; i < NUM_ITERATIONS; ++i) {
       auto errorCode = tfSdk.detectBlink(leftEyeWinkScore, rightEyeWinkScore);
   }
   auto t2 = Clock::now();
   totalTime += std::chrono::duration_cast<std::chrono::milliseconds>(t2 - t1).count();

   std::cout << "Average time blink detection: " << totalTime / NUM_ITERATIONS
             << " ms | " << NUM_ITERATIONS << " iterations" << std::endl;

}

void benchMarkMaskDetection(const std::string& license)
{
    // Initialize sdk
    Trueface::SDK tfSdk;

    bool valid = tfSdk.setLicense(license);

    if (!valid) {
        std::cout << "Error: the provided license is invalid." << std::endl;
        return;
    }

    // Load the image
    Trueface::ErrorCode errorCode = tfSdk.setImage("../images/headshot.jpg");
    if (errorCode != Trueface::ErrorCode::NO_ERROR) {
        std::cout << "Error: could not load the image" << std::endl;
        return;
    }

    Trueface::FaceBoxAndLandmarks faceBoxAndLandmarks;
    bool found;
    tfSdk.detectLargestFace(faceBoxAndLandmarks, found);
    if (!found)
    {
        std::cout<< "Unable to find face in image." << std::endl;
    }

    // Discard first inference
    float maskScore;
    Trueface::MaskLabel maskLabel;
    tfSdk.detectMask(faceBoxAndLandmarks, maskLabel, maskScore);



    const size_t NUM_ITERATIONS = 100;
    double totalTime = 0;

    // Time the creation of the feature vector
    auto t1 = Clock::now();
    for (size_t i = 0; i < NUM_ITERATIONS; ++i) {
        tfSdk.detectMask(faceBoxAndLandmarks, maskLabel, maskScore);
    }
    auto t2 = Clock::now();
    totalTime = std::chrono::duration_cast<std::chrono::milliseconds>(t2 - t1).count();

    std::cout << "Average time mask detection: " << totalTime / NUM_ITERATIONS
                << " ms | " << NUM_ITERATIONS << " iterations" << std::endl;
}